﻿// implementation based on the BLuetooth chat service https://developer.xamarin.com/samples/monodroid/BluetoothChat/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.CompilerServices;
using Android.Bluetooth;
using Android.Util;
using Java.Lang;
using Java.Util;
using Java.Lang.Reflect;

namespace TFM.Droid
{
	/// <summary>
	/// This class does all the work for setting up and managing Bluetooth
	/// connections with other devices. It has a thread that listens for
	/// incoming connections, a thread for connecting with a device, and a
	/// thread for performing data transmissions when connected.
	/// </summary>
	class BluetoothTransferService
	{
		// Debugging
		private const string TAG = "BluetoothTransferService";
		private const bool Debug = true;

		// Unique UUID for this application
		private static UUID MY_UUID = UUID.FromString("3b29c7a5-13b5-42f3-8fb7-8cb80f35aa68");


		// Member fields
		protected BluetoothAdapter _adapter;
		protected ConnectThread connectThread;
		private ConnectedThread connectedThread;
		protected int _state;

		// Constants that indicate the current connection state
		public const int STATE_NONE = 0;       // we're doing nothing
		public const int STATE_LISTEN = 1;     // now listening for incoming connections
		public const int STATE_CONNECTING = 2; // now initiating an outgoing connection
		public const int STATE_CONNECTED = 3;  // now connected to a remote device

		/// <summary>
		/// Constructor. Prepares a new BluetoothTransfer session.
		/// </summary>
		public BluetoothTransferService()
		{
			_adapter = BluetoothAdapter.DefaultAdapter;
			_state = STATE_NONE;
		}

		/// <summary>
		/// Set the current state of the transfer connection.
		/// </summary>
		/// <param name='state'>
		/// An integer defining the current connection state.
		/// </param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		private void SetState(int state)
		{
			if (Debug)
				Log.Debug(TAG, "setState() " + _state + " -> " + state);

			_state = state;
		}

		/// <summary>
		/// Return the current connection state.
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public int GetState()
		{
			return _state;
		}

		// Start the chat service. Specifically start AcceptThread to begin a
		// session in listening (server) mode. Called by the Activity onResume()
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Start()
		{
			if (Debug)
				Log.Debug(TAG, "start");

			// Cancel any thread attempting to make a connection
			if (connectThread != null)
			{
				connectThread.Cancel();
				connectThread = null;
			}

			// Cancel any thread currently running a connection
			if (connectedThread != null)
			{
				connectedThread.Cancel();
				connectedThread = null;
			}

            SetState(STATE_LISTEN);
		}

		/// <summary>
		/// Start the ConnectThread to initiate a connection to a remote device.
		/// </summary>
		/// <param name='device'>
		/// The BluetoothDevice to connect.
		/// </param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Connect(BluetoothDevice device)
		{
			if (Debug)
				Log.Debug(TAG, "connect to: " + device.Name);

			// Cancel any thread attempting to make a connection
			if (_state == STATE_CONNECTING)
			{
				if (connectThread != null)
				{
					connectThread.Cancel();
					connectThread = null;
				}
			}

			// Cancel any thread currently running a connection
			if (connectedThread != null)
			{
				connectedThread.Cancel();
				connectedThread = null;
			}

			// Start the thread to connect with the given device
            SetState(STATE_CONNECTING);
			connectThread = new ConnectThread(device, this);
			connectThread.Start();
		}

		/// <summary>
		/// Start the ConnectedThread to begin managing a Bluetooth connection
		/// </summary>
		/// <param name='socket'>
		/// The BluetoothSocket on which the connection was made.
		/// </param>
		/// <param name='device'>
		/// The BluetoothDevice that has been connected.
		/// </param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Connected(BluetoothSocket socket, BluetoothDevice device)
		{
			if (Debug)
				Log.Debug(TAG, "connected");

			// Cancel the thread that completed the connection
			if (connectThread != null)
			{
				connectThread.Cancel();
				connectThread = null;
			}

			// Cancel any thread currently running a connection
			if (connectedThread != null)
			{
				connectedThread.Cancel();
				connectedThread = null;
			}

			// Start the thread to manage the connection and perform transmissions
			connectedThread = new ConnectedThread(socket, this);
			connectedThread.Start();

			// Send the name of the connected device back to the UI Activity
			SetState(STATE_CONNECTED);
		}

		/// <summary>
		/// Stop all threads.
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Stop()
		{
			if (Debug)
				Log.Debug(TAG, "stop");

			if (connectThread != null)
			{
				connectThread.Cancel();
				connectThread = null;
			}

			if (connectedThread != null)
			{
				connectedThread.Cancel();
				connectedThread = null;
			}


			SetState(STATE_NONE);
		}

		/// <summary>
		/// Write to the ConnectedThread in an unsynchronized manner
		/// </summary>
		/// <param name='out'>
		/// The bytes to write.
		/// </param>
		public byte[] Write(byte[] @out)
		{
			// Create temporary object
			ConnectedThread r;
			// Synchronize a copy of the ConnectedThread
			lock (this)
			{
				r = connectedThread;
			}

			return r.Write(@out);
		}

		/// <summary>
		/// Indicate that the connection attempt failed and notify the UI Activity.
		/// </summary>
		private void ConnectionFailed()
		{
			SetState(STATE_LISTEN);

			// Send a failure message back to the Activity
		}

		/// <summary>
		/// Indicate that the connection was lost and notify the UI Activity.
		/// </summary>
		public void ConnectionLost()
		{
			SetState(STATE_LISTEN);

			// Send a failure message back to the Activity
		}

		/// <summary>
		/// This thread runs while attempting to make an outgoing connection
		/// with a device. It runs straight through; the connection either
		/// succeeds or fails.
		/// </summary>
		protected class ConnectThread : Thread
		{
			private BluetoothSocket mmSocket;
			private BluetoothDevice mmDevice;
			private BluetoothTransferService _service;

			public ConnectThread(BluetoothDevice device, BluetoothTransferService service)
			{
				Log.Info("Connect Thread", "Running");
				mmDevice = device;
				_service = service;
				BluetoothSocket tmp = null;

				// Get a BluetoothSocket for a connection with the
				// given BluetoothDevice
				try
				{
                    tmp = device.CreateRfcommSocketToServiceRecord(MY_UUID);
					Log.Info("created Thread", "Created socket");

				}
                catch (Java.Lang.Exception e)
				{
					Log.Error(TAG, "create() failed", e);
				}
				mmSocket = tmp;

				// After connect, open socket and start connected thread
				// Always cancel discovery because it will slow down a connection
				_service._adapter.CancelDiscovery();

                Thread.Sleep(3000);

				// Make a connection to the BluetoothSocket
				try
				{
					// This is a blocking call and will only return on a
					// successful connection or an exception
					mmSocket.Connect();
				}
				catch (Java.IO.IOException e)
				{

                    try
                    {
                        mmSocket = (BluetoothSocket)device.Class.GetMethod("createRfcommSocket", new Class[] { Java.Lang.Integer.Type }).Invoke(device,1);
                        mmSocket.Connect();

                    }
                    catch (Java.IO.IOException e3)
                    {
                        Log.Error(TAG + "2", e3.Message);
                    }
                    Log.Error(TAG, e.Message);
					_service.ConnectionFailed();
					// Close the socket
					try
					{
						mmSocket.Close();
					}
					catch (Java.IO.IOException e2)
					{
						Log.Error(TAG, "unable to close() socket during connection failure", e2);
					}

					// Start the service over to restart listening mode
					_service.Start();
					return;
				}

				// Reset the ConnectThread because we're done
				lock (this)
				{
					_service.connectThread = null;
				}

				// Start the connected thread
				_service.Connected(mmSocket, mmDevice);
			}

			public override void Run()
			{
				Log.Info(TAG, "BEGIN mConnectThread");
				Name = "ConnectThread";
			}

			public void Cancel()
			{
				try
				{
					mmSocket.Close();
				}
				catch (Java.IO.IOException e)
				{
					Log.Error(TAG, "close() of connect socket failed", e);
				}
			}
		}

		/// <summary>
		/// This thread runs during a connection with a remote device.
		/// It handles all incoming and outgoing transmissions.
		/// </summary>
		private class ConnectedThread : Thread
		{
			private BluetoothSocket mmSocket;
			private Stream mmInStream;
			private Stream mmOutStream;
			private BluetoothTransferService _service;

			public ConnectedThread(BluetoothSocket socket, BluetoothTransferService service)
			{
				Log.Debug(TAG, "create ConnectedThread: ");
				mmSocket = socket;
				_service = service;
				Stream tmpIn = null;
				Stream tmpOut = null;

				// Get the BluetoothSocket input and output streams
				try
				{
					tmpIn = socket.InputStream;
					tmpOut = socket.OutputStream;
				}
				catch (Java.IO.IOException e)
				{
					Log.Error(TAG, "temp sockets not created", e);
				}

				mmInStream = tmpIn;
				mmOutStream = tmpOut;
			}

			public override void Run()
			{
                // Keep listening to the InputStream while connected
				while (true)
				{
					try
					{
						// Do Nothing
					}
					catch (Java.IO.IOException e)
					{
						Log.Error(TAG, "disconnected", e);
						_service.ConnectionLost();
						break;
					}
				}
			}

			/// <summary>
			/// Write to the connected OutStream.
			/// </summary>
			/// <param name='buffer'>
			/// The bytes to write
			/// </param>
			public byte[] Write(byte[] buffer)
			{
				List<byte> responseListBuffer = new List<byte>();
				// Fill empty buffer with error
				responseListBuffer.Add(1);
                responseListBuffer.Add(1);
				//responseListBuffer.AddRange(BitConverter.GetBytes((int)0));
                byte[] responseBuffer = responseListBuffer.ToArray();
                byte[] responseValueBuffer = new byte[1];

				try
				{
                    mmOutStream.Write(buffer, 0, buffer.Length);

                    // Wait for the response
                    mmInStream.Read(responseBuffer, 0, responseBuffer.Length);



                    int leght = responseBuffer[1];
                    List<byte> responseValueListBuffer = new List<byte>();
                    responseValueListBuffer.Add(responseBuffer[0]);
                    for (int i = 0; i < leght; i++)
                    {
                        responseValueListBuffer.Add(0);
                    }

                    responseValueBuffer = responseValueListBuffer.ToArray();
                    mmInStream.Read(responseValueBuffer, 1, leght);
                }
				catch (Java.IO.IOException e)
				{
					Log.Error(TAG, "Exception during write", e);
				}

                return responseValueBuffer;
			}

			public void Cancel()
			{
				try
				{
					mmSocket.Close();
				}
				catch (Java.IO.IOException e)
				{
					Log.Error(TAG, "close() of connect socket failed", e);
				}
			}
		}
	}
}

